@@ -325,19 +325,8 @@ hw_ostc3_transfer (hw_ostc3_device_t *device,
 		}
 	}
 
-	if (delay) {
-		unsigned int count = delay / 100;
-		for (unsigned int i = 0; i < count; ++i) {
-			if (device->available)
-				break;
-
-			size_t available = 0;
-			status = dc_iostream_get_available (device->iostream, &available);
-			if (status == DC_STATUS_SUCCESS && available > 0)
-				break;
-
-			dc_iostream_sleep (device->iostream, 100);
-		}
+	if (delay && device->available == 0) {
+		dc_iostream_poll (device->iostream, delay);
 	}
 
 	if (cmd != EXIT) {
@@ -716,7 +705,6 @@ hw_ostc3_device_foreach (dc_device_t *abstract, dc_dive_callback_t callback, voi
 	// The device maintains an internal counter which is incremented for every
 	// dive, and the current value at the time of the dive is stored in the
 	// dive header. Thus the most recent dive will have the highest value.
-	unsigned int count = 0;
 	unsigned int latest = 0;
 	unsigned int maximum = 0;
 	for (unsigned int i = 0; i < RB_LOGBOOK_COUNT; ++i) {
@@ -732,24 +720,21 @@ hw_ostc3_device_foreach (dc_device_t *abstract, dc_dive_callback_t callback, voi
 			maximum = current;
 			latest = i;
 		}
-
-		count++;
 	}
 
 	// Calculate the total and maximum size.
 	unsigned int ndives = 0;
 	unsigned int size = 0;
 	unsigned int maxsize = 0;
-	for (unsigned int i = 0; i < count; ++i) {
+	unsigned char dive[RB_LOGBOOK_COUNT] = {0};
+	for (unsigned int i = 0; i < RB_LOGBOOK_COUNT; ++i) {
 		unsigned int idx = (latest + RB_LOGBOOK_COUNT - i) % RB_LOGBOOK_COUNT;
 		unsigned int offset = idx * logbook->size;
 
-		// Uninitialized header entries should no longer be present at this
-		// stage, unless the dives are interleaved with empty entries. But
-		// that's something we don't support at all.
+		// Ignore uninitialized header entries.
 		if (array_isequal (header + offset, logbook->size, 0xFF)) {
 			WARNING (abstract->context, "Unexpected empty header found.");
-			break;
+			continue;
 		}
 
 		// Calculate the profile length.
@@ -773,6 +758,7 @@ hw_ostc3_device_foreach (dc_device_t *abstract, dc_dive_callback_t callback, voi
 		if (length > maxsize)
 			maxsize = length;
 		size += length;
+		dive[ndives] = idx;
 		ndives++;
 	}
 
@@ -796,7 +782,7 @@ hw_ostc3_device_foreach (dc_device_t *abstract, dc_dive_callback_t callback, voi
 
 	// Download the dives.
 	for (unsigned int i = 0; i < ndives; ++i) {
-		unsigned int idx = (latest + RB_LOGBOOK_COUNT - i) % RB_LOGBOOK_COUNT;
+		unsigned int idx = dive[i];
 		unsigned int offset = idx * logbook->size;
 
 		// Calculate the profile length.
