@@ -35,6 +35,7 @@
 #include "context-private.h"
 #include "parser-private.h"
 #include "array.h"
+#include "field-cache.h"
 
 #define ISINSTANCE(parser)	( \
 	dc_parser_isinstance((parser), &shearwater_predator_parser_vtable) || \
@@ -58,6 +59,7 @@
 #define LOG_RECORD_CLOSING_5       0x25
 #define LOG_RECORD_CLOSING_6       0x26
 #define LOG_RECORD_CLOSING_7       0x27
+#define LOG_RECORD_INFO_EVENT      0x30
 #define LOG_RECORD_FINAL           0xFF
 
 #define SZ_BLOCK   0x80
@@ -76,15 +78,30 @@
 
 #define NGASMIXES 10
 #define MAXSTRINGS 32
-#define NRECORDS  7
+#define NTANKS    2
+#define NRECORDS  8
 
 #define PREDATOR 2
 #define PETREL   3
 
+#define INFO_EVENT_TAG_LOG 38
+
 #define UNDEFINED 0xFFFFFFFF
 
 typedef struct shearwater_predator_parser_t shearwater_predator_parser_t;
 
+typedef struct shearwater_predator_gasmix_t {
+	unsigned int oxygen;
+	unsigned int helium;
+} shearwater_predator_gasmix_t;
+
+typedef struct shearwater_predator_tank_t {
+	unsigned int enabled;
+	unsigned int beginpressure;
+	unsigned int endpressure;
+	unsigned int battery;
+} shearwater_predator_tank_t;
+
 struct shearwater_predator_parser_t {
 	dc_parser_t base;
 	unsigned int model;
@@ -100,18 +117,19 @@ struct shearwater_predator_parser_t {
 	unsigned int closing[NRECORDS];
 	unsigned int final;
 	unsigned int ngasmixes;
-	unsigned int oxygen[NGASMIXES];
-	unsigned int helium[NGASMIXES];
+	unsigned int ntanks;
+	shearwater_predator_gasmix_t gasmix[NGASMIXES];
+	shearwater_predator_tank_t tank[NTANKS];
+	unsigned int tankidx[NTANKS];
 	unsigned int calibrated;
 	double calibration[3];
 	unsigned int serial;
-	dc_divemode_t mode;
 	unsigned int units;
 	unsigned int atmospheric;
 	unsigned int density;
 
-	/* String fields */
-	dc_field_string_t strings[MAXSTRINGS];
+	/* Generic field cache */
+	struct dc_field_cache cache;
 };
 
 static dc_status_t shearwater_predator_parser_set_data (dc_parser_t *abstract, const unsigned char *data, unsigned int size);
@@ -147,7 +165,7 @@ shearwater_predator_find_gasmix (shearwater_predator_parser_t *parser, unsigned
 {
 	unsigned int i = 0;
 	while (i < parser->ngasmixes) {
-		if (o2 == parser->oxygen[i] && he == parser->helium[i])
+		if (o2 == parser->gasmix[i].oxygen && he == parser->gasmix[i].helium)
 			break;
 		i++;
 	}
@@ -200,18 +218,27 @@ shearwater_common_parser_create (dc_parser_t **out, dc_context_t *context, unsig
 	parser->final = UNDEFINED;
 	parser->ngasmixes = 0;
 	for (unsigned int i = 0; i < NGASMIXES; ++i) {
-		parser->oxygen[i] = 0;
-		parser->helium[i] = 0;
+		parser->gasmix[i].oxygen = 0;
+		parser->gasmix[i].helium = 0;
+	}
+	parser->ntanks = 0;
+	for (unsigned int i = 0; i < NTANKS; ++i) {
+		parser->tank[i].enabled = 0;
+		parser->tank[i].beginpressure = 0;
+		parser->tank[i].endpressure = 0;
+		parser->tank[i].battery = 0;
+		parser->tankidx[i] = i;
 	}
 	parser->calibrated = 0;
 	for (unsigned int i = 0; i < 3; ++i) {
 		parser->calibration[i] = 0.0;
 	}
-	parser->mode = DC_DIVEMODE_OC;
 	parser->units = METRIC;
 	parser->density = 1025;
 	parser->atmospheric = ATM / (BAR / 1000);
 
+	DC_ASSIGN_FIELD(parser->cache, DIVEMODE, DC_DIVEMODE_OC);
+
 	*out = (dc_parser_t *) parser;
 
 	return DC_STATUS_SUCCESS;
@@ -250,18 +277,26 @@ shearwater_predator_parser_set_data (dc_parser_t *abstract, const unsigned char
 	parser->final = UNDEFINED;
 	parser->ngasmixes = 0;
 	for (unsigned int i = 0; i < NGASMIXES; ++i) {
-		parser->oxygen[i] = 0;
-		parser->helium[i] = 0;
+		parser->gasmix[i].oxygen = 0;
+		parser->gasmix[i].helium = 0;
+	}
+	parser->ntanks = 0;
+	for (unsigned int i = 0; i < NTANKS; ++i) {
+		parser->tank[i].enabled = 0;
+		parser->tank[i].beginpressure = 0;
+		parser->tank[i].endpressure = 0;
+		parser->tankidx[i] = i;
 	}
 	parser->calibrated = 0;
 	for (unsigned int i = 0; i < 3; ++i) {
 		parser->calibration[i] = 0.0;
 	}
-	parser->mode = DC_DIVEMODE_OC;
 	parser->units = METRIC;
 	parser->density = 1025;
 	parser->atmospheric = ATM / (BAR / 1000);
 
+	DC_ASSIGN_FIELD(parser->cache, DIVEMODE, DC_DIVEMODE_OC);
+
 	return DC_STATUS_SUCCESS;
 }
 
@@ -287,74 +322,6 @@ shearwater_predator_parser_get_datetime (dc_parser_t *abstract, dc_datetime_t *d
 	return DC_STATUS_SUCCESS;
 }
 
-/*
- * These string cache interfaces should be some generic
- * library rather than copied for all the dive computers.
- *
- * This is just copied from the EON Steel code.
- */
-static void
-add_string(shearwater_predator_parser_t *parser, const char *desc, const char *value)
-{
-	int i;
-
-	for (i = 0; i < MAXSTRINGS; i++) {
-		dc_field_string_t *str = parser->strings+i;
-		if (str->desc)
-			continue;
-		str->desc = desc;
-		str->value = strdup(value);
-		break;
-	}
-}
-
-static void
-add_string_fmt(shearwater_predator_parser_t *parser, const char *desc, const char *fmt, ...)
-{
-	char buffer[256];
-	va_list ap;
-
-	/*
-	 * We ignore the return value from vsnprintf, and we
-	 * always NUL-terminate the destination buffer ourselves.
-	 *
-	 * That way we don't have to worry about random bad legacy
-	 * implementations.
-	 */
-	va_start(ap, fmt);
-	buffer[sizeof(buffer)-1] = 0;
-	(void) vsnprintf(buffer, sizeof(buffer)-1, fmt, ap);
-	va_end(ap);
-
-	return add_string(parser, desc, buffer);
-}
-
-// The Battery state is a big-endian word:
-//
-//  ffff = not paired / no comms for 90 s
-//  fffe = no comms for 30 s
-//
-// Otherwise:
-//   - top four bits are battery state (0 - normal, 1 - critical, 2 - warning)
-//   - bottom 12 bits are pressure in 2 psi increments (0..8k psi)
-//
-// This returns the state as a bitmask (so you can see all states it had
-// during the dive). Note that we currently do not report pairing and
-// communication lapses. Todo?
-static unsigned int
-battery_state(const unsigned char *data)
-{
-	unsigned int pressure = array_uint16_be(data);
-	unsigned int state;
-
-	if ((pressure & 0xFFF0) == 0xFFF0)
-		return 0;
-	state = pressure >> 12;
-	if (state > 2)
-		return 0;
-	return 1u << state;
-}
-
 // Show the battery state
 //
 // NOTE! Right now it only shows the most serious bit
@@ -367,6 +334,8 @@ battery_state(const unsigned char *data)
 static void
 add_battery_info(shearwater_predator_parser_t *parser, const char *desc, unsigned int state)
 {
+	// We don't know what other state bits than 0-2 mean
+	state &= 7;
 	if (state >= 1 && state <= 7) {
 		static const char *states[8] = {
 			"",		// 000 - No state bits, not used
@@ -378,7 +347,7 @@ add_battery_info(shearwater_predator_parser_t *parser, const char *desc, unsigne
 			"critical",	// 110 - warning and critical
 			"critical",	// 111 - normal, warning and critical
 		};
-		add_string(parser, desc, states[state]);
+		dc_field_add_string(&parser->cache, desc, states[state]);
 	}
 }
 
@@ -390,16 +359,16 @@ add_deco_model(shearwater_predator_parser_t *parser, const unsigned char *data)
 
 	switch	(data[idx_deco_model]) {
 	case 0:
-		add_string_fmt(parser, "Deco model", "GF %u/%u", data[4], data[5]);
+		dc_field_add_string_fmt(&parser->cache, "Deco model", "GF %u/%u", data[4], data[5]);
 		break;
 	case 1:
-		add_string_fmt(parser, "Deco model", "VPM-B +%u", data[idx_deco_model + 1]);
+		dc_field_add_string_fmt(&parser->cache, "Deco model", "VPM-B +%u", data[idx_deco_model + 1]);
 		break;
 	case 2:
-		add_string_fmt(parser, "Deco model", "VPM-B/GFS +%u %u%%", data[idx_deco_model + 1], data[idx_gfs]);
+		dc_field_add_string_fmt(&parser->cache, "Deco model", "VPM-B/GFS +%u %u%%", data[idx_deco_model + 1], data[idx_gfs]);
 		break;
 	default:
-		add_string_fmt(parser, "Deco model", "Unknown model %d", data[idx_deco_model]);
+		dc_field_add_string_fmt(&parser->cache, "Deco model", "Unknown model %d", data[idx_deco_model]);
 	}
 }
 
@@ -412,22 +381,22 @@ add_battery_type(shearwater_predator_parser_t *parser, const unsigned char *data
 	unsigned int idx_battery_type = parser->pnf ? parser->opening[4] + 9 : 120;
 	switch (data[idx_battery_type]) {
 	case 1:
-		add_string(parser, "Battery type", "1.5V Alkaline");
+		dc_field_add_string(&parser->cache, "Battery type", "1.5V Alkaline");
 		break;
 	case 2:
-		add_string(parser, "Battery type", "1.5V Lithium");
+		dc_field_add_string(&parser->cache, "Battery type", "1.5V Lithium");
 		break;
 	case 3:
-		add_string(parser, "Battery type", "1.2V NiMH");
+		dc_field_add_string(&parser->cache, "Battery type", "1.2V NiMH");
 		break;
 	case 4:
-		add_string(parser, "Battery type", "3.6V Saft");
+		dc_field_add_string(&parser->cache, "Battery type", "3.6V Saft");
 		break;
 	case 5:
-		add_string(parser, "Battery type", "3.7V Li-Ion");
+		dc_field_add_string(&parser->cache, "Battery type", "3.7V Li-Ion");
 		break;
 	default:
-		add_string_fmt(parser, "Battery type", "unknown type %d", data[idx_battery_type]);
+		dc_field_add_string_fmt(&parser->cache, "Battery type", "unknown type %d", data[idx_battery_type]);
 		break;
 	}
 }
@@ -442,7 +411,11 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 	if (parser->cached) {
 		return DC_STATUS_SUCCESS;
 	}
-	memset(parser->strings, 0, sizeof(parser->strings));
+	memset(&parser->cache, 0, sizeof(parser->cache));
+
+	// Log versions before 6 weren't reliably stored in the data, but
+	// 6 is also the oldest version that we assume in our code
+	unsigned int logversion = 0;
 
 	// Verify the minimum length.
 	if (size < 2) {
@@ -486,6 +459,9 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 			parser->opening[i] = 0;
 			parser->closing[i] = size - footersize;
 		}
+
+		// Log version
+		logversion = data[127];
 	}
 
 	// Default dive mode.
@@ -493,13 +469,10 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 
 	// Get the gas mixes.
 	unsigned int ngasmixes = 0;
-	unsigned int oxygen[NGASMIXES] = {0};
-	unsigned int helium[NGASMIXES] = {0};
+	shearwater_predator_gasmix_t gasmix[NGASMIXES] = {0};
+	shearwater_predator_tank_t tank[NTANKS] = {0};
 	unsigned int o2_previous = 0, he_previous = 0;
 
-	// Transmitter battery levels
-	unsigned int t1_battery = 0, t2_battery = 0;
-
 	unsigned int offset = headersize;
 	unsigned int length = size - footersize;
 	while (offset + parser->samplesize <= length) {
@@ -526,7 +499,7 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 				// Find the gasmix in the list.
 				unsigned int idx = 0;
 				while (idx < ngasmixes) {
-					if (o2 == oxygen[idx] && he == helium[idx])
+					if (o2 == gasmix[idx].oxygen && he == gasmix[idx].helium)
 						break;
 					idx++;
 				}
@@ -537,8 +510,8 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 						ERROR (abstract->context, "Maximum number of gas mixes reached.");
 						return DC_STATUS_NOMEMORY;
 					}
-					oxygen[idx] = o2;
-					helium[idx] = he;
+					gasmix[idx].oxygen = o2;
+					gasmix[idx].helium = he;
 					ngasmixes = idx + 1;
 				}
 
@@ -546,16 +519,44 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 				he_previous = he;
 			}
 
-			// Transmitter battery levels
-			// T1 at offset 27, T2 at offset 19
-			t1_battery |= battery_state(data + offset + 27 + pnf);
-			t2_battery |= battery_state(data + offset + 19 + pnf);
+			// Tank pressure
+			if (logversion >= 7) {
+				const unsigned int idx[NTANKS] = {27, 19};
+				for (unsigned int i = 0; i < NTANKS; ++i) {
+					// Values above 0xFFF0 are special codes:
+					//    0xFFFF AI is off
+					//    0xFFFE No comms for 90 seconds+
+					//    0xFFFD No comms for 30 seconds
+					//    0xFFFC Transmitter not paired
+					// For regular values, the top 4 bits contain the battery
+					// level (0=normal, 1=critical, 2=warning), and the lower 12
+					// bits the tank pressure in units of 2 psi.
+					unsigned int pressure = array_uint16_be (data + offset + pnf + idx[i]);
+					if (pressure < 0xFFF0) {
+						unsigned int battery = 1u << (pressure >> 12);
+						pressure &= 0x0FFF;
+						if (!tank[i].enabled) {
+							tank[i].enabled = 1;
+							tank[i].beginpressure = pressure;
+							tank[i].endpressure = pressure;
+							tank[i].battery = 0;
+						}
+						tank[i].endpressure = pressure;
+						tank[i].battery |= battery;
+					}
+				}
+			}
 		} else if (type == LOG_RECORD_FREEDIVE_SAMPLE) {
 			// Freedive record
 			mode = DC_DIVEMODE_FREEDIVE;
 		} else if (type >= LOG_RECORD_OPENING_0 && type <= LOG_RECORD_OPENING_7) {
 			// Opening record
 			parser->opening[type - LOG_RECORD_OPENING_0] = offset;
+
+			// Log version
+			if (type == LOG_RECORD_OPENING_4) {
+				logversion = data[offset + 16];
+			}
 		} else if (type >= LOG_RECORD_CLOSING_0 && type <= LOG_RECORD_CLOSING_7) {
 			// Closing record
 			parser->closing[type - LOG_RECORD_CLOSING_0] = offset;
@@ -575,16 +576,7 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 		}
 	}
 
-	// Log versions before 6 weren't reliably stored in the data, but
-	// 6 is also the oldest version that we assume in our code
-	unsigned int logversion = data[parser->opening[4] + (pnf ? 16 : 127)];
-	add_string_fmt(parser, "Logversion", "%d%s", logversion, pnf ? "(PNF)" : "");
-
-	// The transmitter battery levels are only valid for logversion 7+
-	if (logversion < 7) {
-		t1_battery = 0;
-		t2_battery = 0;
-	}
+	dc_field_add_string_fmt(&parser->cache, "Logversion", "%d%s", logversion, pnf ? "(PNF)" : "");
 
 	// Cache sensor calibration for later use
 	unsigned int nsensors = 0, ndefaults = 0;
@@ -615,11 +607,11 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 		WARNING (abstract->context, "Disabled all O2 sensors due to a default calibration value.");
 		parser->calibrated = 0;
 		if (mode != DC_DIVEMODE_OC)
-			add_string(parser, "PPO2 source", "voted/averaged");
+			dc_field_add_string(&parser->cache, "PPO2 source", "voted/averaged");
 	} else {
 		parser->calibrated = data[base];
 		if (mode != DC_DIVEMODE_OC)
-			add_string(parser, "PPO2 source", "cells");
+			dc_field_add_string(&parser->cache, "PPO2 source", "cells");
 	}
 
 	// Cache the data for later use.
@@ -629,23 +621,33 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 	parser->footersize = footersize;
 	parser->ngasmixes = ngasmixes;
 	for (unsigned int i = 0; i < ngasmixes; ++i) {
-		parser->oxygen[i] = oxygen[i];
-		parser->helium[i] = helium[i];
+		parser->gasmix[i] = gasmix[i];
+	}
+	parser->ntanks = 0;
+	for (unsigned int i = 0; i < NTANKS; ++i) {
+		if (tank[i].enabled) {
+			parser->tankidx[i] = parser->ntanks;
+			parser->tank[parser->ntanks] = tank[i];
+			parser->ntanks++;
+		} else {
+			parser->tankidx[i] = UNDEFINED;
+		}
 	}
-	parser->mode = mode;
 	parser->units = data[parser->opening[0] + 8];
 	parser->atmospheric = array_uint16_be (data + parser->opening[1] + (parser->pnf ? 16 : 47));
 	parser->density = array_uint16_be (data + parser->opening[3] + (parser->pnf ? 3 : 83));
 	parser->cached = 1;
 
-	add_string_fmt(parser, "Serial", "%08x", parser->serial);
+	DC_ASSIGN_FIELD(parser->cache, DIVEMODE, mode);
+
+	dc_field_add_string_fmt(&parser->cache, "Serial", "%08x", parser->serial);
 	// bytes 1-31 are identical in all formats
-	add_string_fmt(parser, "FW Version", "%2x", data[19]);
+	dc_field_add_string_fmt(&parser->cache, "FW Version", "%2x", data[19]);
 	add_deco_model(parser, data);
 	add_battery_type(parser, data);
-	add_string_fmt(parser, "Battery at end", "%.1f V", data[9] / 10.0);
-	add_battery_info(parser, "T1 battery", t1_battery);
-	add_battery_info(parser, "T2 battery", t2_battery);
+	dc_field_add_string_fmt(&parser->cache, "Battery at end", "%.1f V", data[9] / 10.0);
+	add_battery_info(parser, "T1 battery", tank[0].battery);
+	add_battery_info(parser, "T2 battery", tank[1].battery);
 
 	return DC_STATUS_SUCCESS;
 }
@@ -663,6 +665,7 @@ shearwater_predator_parser_get_field (dc_parser_t *abstract, dc_field_type_t typ
 		return rc;
 
 	dc_gasmix_t *gasmix = (dc_gasmix_t *) value;
+	dc_tank_t *tank = (dc_tank_t *) value;
 	dc_salinity_t *water = (dc_salinity_t *) value;
 	dc_field_string_t *string = (dc_field_string_t *) value;
 
@@ -686,10 +689,21 @@ shearwater_predator_parser_get_field (dc_parser_t *abstract, dc_field_type_t typ
 			*((unsigned int *) value) = parser->ngasmixes;
 			break;
 		case DC_FIELD_GASMIX:
-			gasmix->oxygen = parser->oxygen[flags] / 100.0;
-			gasmix->helium = parser->helium[flags] / 100.0;
+			gasmix->oxygen = parser->gasmix[flags].oxygen / 100.0;
+			gasmix->helium = parser->gasmix[flags].helium / 100.0;
 			gasmix->nitrogen = 1.0 - gasmix->oxygen - gasmix->helium;
 			break;
+		case DC_FIELD_TANK_COUNT:
+			*((unsigned int *) value) = parser->ntanks;
+			break;
+		case DC_FIELD_TANK:
+			tank->type = DC_TANKVOLUME_NONE;
+			tank->volume = 0.0;
+			tank->workpressure = 0.0;
+			tank->beginpressure = parser->tank[flags].beginpressure * 2 * PSI / BAR;
+			tank->endpressure   = parser->tank[flags].endpressure   * 2 * PSI / BAR;
+			tank->gasmix = DC_GASMIX_UNKNOWN;
+			break;
 		case DC_FIELD_SALINITY:
 			if (parser->density == 1000)
 				water->type = DC_WATER_FRESH;
@@ -701,17 +715,9 @@ shearwater_predator_parser_get_field (dc_parser_t *abstract, dc_field_type_t typ
 			*((double *) value) = parser->atmospheric / 1000.0;
 			break;
 		case DC_FIELD_DIVEMODE:
-			*((dc_divemode_t *) value) = parser->mode;
-			break;
+			return DC_FIELD_VALUE(parser->cache, value, DIVEMODE);
 		case DC_FIELD_STRING:
-			if (flags < MAXSTRINGS) {
-				dc_field_string_t *p = parser->strings + flags;
-				if (p->desc) {
-					*string = *p;
-					break;
-				}
-			}
-			return DC_STATUS_UNSUPPORTED;
+			return dc_field_get_string(&parser->cache, flags, string);
 		default:
 			return DC_STATUS_UNSUPPORTED;
 		}
@@ -759,6 +765,32 @@ shearwater_predator_parser_samples_foreach (dc_parser_t *abstract, dc_sample_cal
 		if (pnf && data[offset] == LOG_RECORD_FINAL && data[offset + 1] == 0xFD)
 			break;
 
+		if (pnf && data[offset] == LOG_RECORD_INFO_EVENT) {
+			// additional events defined in PNF
+			INFO(abstract->context, "PNF INFO_EVENT ID %u time %u W1 %u W2 %u", data[offset + 1],
+											    array_uint32_be (data + offset + 4),
+											    array_uint32_be (data + offset + 8),
+											    array_uint32_be (data + offset + 12));
+			if (data[offset + 1] == INFO_EVENT_TAG_LOG) {
+				// this is a TAG
+				// its time is a unix timestamp, so we need to subtract the dive start time
+				unsigned int tag_time = array_uint32_be (data + offset + 4) - array_uint32_be (data + parser->opening[0] + 12);
+				unsigned int tag_heading = array_uint32_be (data + offset + 8);
+				unsigned int tag_type = array_uint32_be (data + offset + 12);
+				// heading is only valid if 0..360
+				// type is only valid if 0..5
+				if (tag_heading <= 360 && tag_type <= 5) {
+					// encode this as a bookmark event, using the flags to capture the type and value for heading
+					sample.event.type = SAMPLE_EVENT_BOOKMARK;
+					sample.event.time = tag_time;
+					sample.event.flags = (tag_type + 1) << SAMPLE_FLAGS_TYPE_SHIFT; // 0 means it isn't a tag
+					sample.event.value = tag_heading;
+					if (callback) callback (DC_SAMPLE_EVENT, sample, userdata);
+				}
+			}
+			offset += parser->samplesize;
+			continue;
+		}
 		// Ignore blocks that aren't dive samples
 		if (pnf && data[offset] != LOG_RECORD_DIVE_SAMPLE) {
 			offset += parser->samplesize;
@@ -808,19 +840,19 @@ shearwater_predator_parser_samples_foreach (dc_parser_t *abstract, dc_sample_cal
 			if ((status & OC) == 0) {
 				// PPO2
 				if ((status & PPO2_EXTERNAL) == 0) {
-#ifdef SENSOR_AVERAGE
-					sample.ppo2 = data[offset + pnf + 6] / 100.0;
-					if (callback) callback (DC_SAMPLE_PPO2, sample, userdata);
-#else
-					sample.ppo2 = data[offset + pnf + 12] * parser->calibration[0];
-					if (callback && (parser->calibrated & 0x01)) callback (DC_SAMPLE_PPO2, sample, userdata);
+					if (!parser->calibrated) {
+						sample.ppo2 = data[offset + pnf + 6] / 100.0;
+						if (callback) callback (DC_SAMPLE_PPO2, sample, userdata);
+					} else {
+						sample.ppo2 = data[offset + pnf + 12] * parser->calibration[0];
+						if (callback && (parser->calibrated & 0x01)) callback (DC_SAMPLE_PPO2, sample, userdata);
 
-					sample.ppo2 = data[offset + pnf + 14] * parser->calibration[1];
-					if (callback && (parser->calibrated & 0x02)) callback (DC_SAMPLE_PPO2, sample, userdata);
+						sample.ppo2 = data[offset + pnf + 14] * parser->calibration[1];
+						if (callback && (parser->calibrated & 0x02)) callback (DC_SAMPLE_PPO2, sample, userdata);
 
-					sample.ppo2 = data[offset + pnf + 15] * parser->calibration[2];
-					if (callback && (parser->calibrated & 0x04)) callback (DC_SAMPLE_PPO2, sample, userdata);
-#endif
+						sample.ppo2 = data[offset + pnf + 15] * parser->calibration[2];
+						if (callback && (parser->calibrated & 0x04)) callback (DC_SAMPLE_PPO2, sample, userdata);
+					}
 				}
 
 				// Setpoint
@@ -877,28 +909,24 @@ shearwater_predator_parser_samples_foreach (dc_parser_t *abstract, dc_sample_cal
 			// for logversion 7 and newer (introduced for Perdix AI)
 			// detect tank pressure
 			if (parser->logversion >= 7) {
-				// Tank pressure
-				// Values above 0xFFF0 are special codes:
-				//    0xFFFF AI is off
-				//    0xFFFE No comms for 90 seconds+
-				//    0xFFFD No comms for 30 seconds
-				//    0xFFFC Transmitter not paired
-				// For regular values, the top 4 bits contain the battery
-				// level (0=normal, 1=critical, 2=warning), and the lower 12
-				// bits the tank pressure in units of 2 psi.
-				unsigned int pressure = array_uint16_be (data + offset + pnf + 27);
-				if (pressure < 0xFFF0) {
-					pressure &= 0x0FFF;
-					sample.pressure.tank = 0;
-					sample.pressure.value = pressure * 2 * PSI / BAR;
-					if (callback) callback (DC_SAMPLE_PRESSURE, sample, userdata);
-				}
-				pressure = array_uint16_be (data + offset + pnf + 19);
-				if (pressure < 0xFFF0) {
-					pressure &= 0x0FFF;
-					sample.pressure.tank = 1;
-					sample.pressure.value = pressure * 2 * PSI / BAR;
-					if (callback) callback (DC_SAMPLE_PRESSURE, sample, userdata);
+				const unsigned int idx[NTANKS] = {27, 19};
+				for (unsigned int i = 0; i < NTANKS; ++i) {
+					// Tank pressure
+					// Values above 0xFFF0 are special codes:
+					//    0xFFFF AI is off
+					//    0xFFFE No comms for 90 seconds+
+					//    0xFFFD No comms for 30 seconds
+					//    0xFFFC Transmitter not paired
+					// For regular values, the top 4 bits contain the battery
+					// level (0=normal, 1=critical, 2=warning), and the lower 12
+					// bits the tank pressure in units of 2 psi.
+					unsigned int pressure = array_uint16_be (data + offset + pnf + idx[i]);
+					if (pressure < 0xFFF0) {
+						pressure &= 0x0FFF;
+						sample.pressure.tank = parser->tankidx[i];
+						sample.pressure.value = pressure * 2 * PSI / BAR;
+						if (callback) callback (DC_SAMPLE_PRESSURE, sample, userdata);
+					}
 				}
 
 				// Gas time remaining in minutes
